#include "application.hpp"
#include "board.hpp"
#include "display_utils.hpp"
#include "utils/delay.hpp"
#include "player.hpp"

using namespace Voltmeter;


Application::Application()
    : reader(player)
    , timeMs(0)
    , OnOffButton(LongPressTimeMs)
    , RangeButton(LongPressTimeMs)
    , TalkButton(LongPressTimeMs, true)
{}


void Application::Main()
{
    Board::PreInit();
    Utils::DelayMs(LongPressTimeMs);
    Board::Init();

    DisplayUtils::LoadChars();
    
    configuration.Read();
    measurer.Configure(Configuration::SamplesN);
    measurer.SetRange(configuration.range);
    isBacklightOn = configuration.backlight;
    Board::Display::Driver::SetBacklight(isBacklightOn);
    volumeDivider = VolumeToVolumeDivider(configuration.volume);

    sei();

    Board::SetTimeout(SdInitTimeout);
    
    if (!player.MountFs())
    {
        ErrorP(PSTR("Brak karty SD."), true);
    }

    Board::SetTimeout(MainLoopTimeout);

    while (true)
    {
        Board::Nop();

        player.Play();
        reader.Execute();

        this->Measure();

        if (isBacklightOn && !configuration.backlight && (this->GetTimeMs() - backlightOnTime > BacklightTimeMs))
        {
            isBacklightOn = false;
            Board::Display::Driver::SetBacklight(false);
        }

        this->GetInput();
    }
}


void Application::PowerOff()
{
    cli();
    Board::SetTimeout(ConfigWriteTimeout);

    Board::Display::Clear();
    Board::Display::Driver::SetBacklight(false);
    configuration.Write();
    
    // Wait until button is released
    Board::WaitForPin<Board::OnOffButtonPin>(true);
    
    Utils::DelayMs(100);
    Board::PowerEnablePin::SetLow();

    Board::Loop();
}


void Application::ErrorP(const char * message, bool isFatal)
{
    cli();
    Board::Display::Init();
    Board::Display::Driver::SetBacklight(true);
    Board::Display::Clear();
    Board::Display::PutStringP(PSTR("Blad:"));
    Board::Display::Move(0, 1);
    Board::Display::PutStringP(message);

    Utils::DelayMs(2000);

    if (isFatal)
    {
        // Wait for button release
        Board::WaitForPin<Board::OnOffButtonPin>(true);
        Board::WaitForPin<Board::OnOffButtonPin>(false);
        this->PowerOff();
    }

    sei();
    Board::Display::Driver::SetBacklight(isBacklightOn);
}


int32_t Application::GetTimeMs()
{
    int32_t result = 0;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        result = timeMs;
    }

    return result;
}


void Application::Measure()
{
    if (measurer.HasMeasurement())
    {
        Measurement measurement = measurer.GetMeasurement();
        uint32_t calibration = configuration.GetCalibration(measurement.range);
        dcVoltage = measurement.GetDcVoltage(Configuration::DigitsN, calibration);
        acVoltage = measurement.GetAcVoltage(Configuration::DigitsN);

        reader.ReadMeasurement(acVoltage, dcVoltage);

        if (configuration.autoRange)
        {
            Voltage maxVoltage = measurement.GetMaxVoltage(Configuration::DigitsN, calibration);

            if (maxVoltage.IsOverRange() && measurement.range != Range::Max)
            {
                configuration.range = measurement.range + 1;
                measurer.SetRange(configuration.range);
            }
            else if (maxVoltage.IsUnderRange() && measurement.range != Range::Min)
            {
                configuration.range = measurement.range - 1;
                measurer.SetRange(configuration.range);
            }
        }
        
        this->UpdateDisplay();
    }
}


void Application::GetInput()
{
    int32_t timeMs = this->GetTimeMs();
    bool updateDisplay = false;

    if (!Board::RangeButtonPin::Get() && !Board::TalkButtonPin::Get())
    {
        this->Calibrate();
        return;
    }

    OnOffButton.SetState(!Board::OnOffButtonPin::Get(), timeMs);
    RangeButton.SetState(!Board::RangeButtonPin::Get(), timeMs);
    TalkButton.SetState(!Board::TalkButtonPin::Get(), timeMs);

    if (OnOffButton.WasPressed())
    {
        if (configuration.backlight)
        {
            configuration.backlight = false;
            isBacklightOn = false;
            Board::Display::Driver::SetBacklight(false);
            updateDisplay = true;
        }
        else if (isBacklightOn)
        {
            configuration.backlight = true;
            updateDisplay = true;
        }
        else
        {
            isBacklightOn = true;
            backlightOnTime = timeMs;
            Board::Display::Driver::SetBacklight(true);
        }
    }

    if (OnOffButton.WasLongPressed())
    {
        this->PowerOff();
    }

    if (RangeButton.WasPressed())
    {
        if (configuration.autoRange)
        {
            configuration.autoRange = false;
        }
        else
        {
            ++configuration.range;
        }

        measurer.SetRange(configuration.range);
        reader.ReadRange(configuration.range);
        updateDisplay = true;
    }

    if (RangeButton.WasLongPressed())
    {
        configuration.autoRange = true;
        reader.ReadAutoRange();
    }

    if (TalkButton.WasPressed())
    {
        reader.ForceReadMeasurement(acVoltage, dcVoltage);
    }

    if (TalkButton.WasLongPressed())
    {
        this->ChangeVolume();
        updateDisplay = true;
    }

    if (updateDisplay)
    {
        this->UpdateDisplay();
    }
}


uint8_t Application::VolumeToVolumeDivider(uint8_t volume)
{
    switch (configuration.volume)
    {
        case 0:
        default:
            return VolumeLevel0;

        case 1:
            return VolumeLevel1;

        case 2:
            return VolumeLevel2;

        case 3:
            return VolumeLevel3;
    }
}


void Application::ChangeVolume()
{
    ++configuration.volume;

    if (configuration.volume >= 4)
    {
        configuration.volume = 0;
    }

    volumeDivider = VolumeToVolumeDivider(configuration.volume);
}


void Application::UpdateDisplay()
{
    uint8_t voltageX = 5 - Configuration::DigitsN / 2;

    DisplayUtils::LoadVolumeChar(configuration.volume);

    if (acVoltage.value == 0)
    {
        DisplayUtils::LoadWaveFormChar(WaveForm::Dc);
    }
    else
    {
        if (dcVoltage.value == 0)
        {
            DisplayUtils::LoadWaveFormChar(WaveForm::Ac);
        }
        else
        {
            DisplayUtils::LoadWaveFormChar(WaveForm::AcDc);
        }
    }

    Board::Display::Move(voltageX, 0);
    DisplayUtils::DisplayVoltage(dcVoltage);
    Board::Display::Pad(' ', 2);
    Board::Display::Move(voltageX, 1);
    DisplayUtils::DisplayVoltage(acVoltage);
    Board::Display::Pad(' ', 2);

    Board::Display::Move(voltageX - 2, 0);
    Board::Display::PutChar(DisplayUtils::CharWaveForm);

    Board::Display::Move(14, 0);
    Board::Display::PutChar(DisplayUtils::CharSpeaker);
    Board::Display::PutChar(DisplayUtils::CharVolume);

    Board::Display::Move(13, 1);

    if (configuration.GetCalibration(Range::_400mV) == 0)
    {
        Board::Display::PutChar(DisplayUtils::CharWarning);
    }
    else
    {
        Board::Display::PutChar(' ');
    }

    Board::Display::PutChar(configuration.backlight ? DisplayUtils::CharBacklightLock : ' ');
    Board::Display::PutChar(configuration.autoRange ? DisplayUtils::CharAutoRange : ' ');
}


void Application::CalibrateRange(Range const range)
{
    measurer.SetRange(range);
    Measurement measurement;

    for (uint8_t i = 0; i < 3; ++i)
    {
        while (!measurer.HasMeasurement())
        {
            Board::Nop();
        }

        measurement = measurer.GetMeasurement();
    }

    Board::Display::PutChar(0xff);

    uint32_t calibration = 0;

    for (uint8_t i = 0; i < 10; ++i)
    {
        while (!measurer.HasMeasurement())
        {
            Board::Nop();
        }

        measurement = measurer.GetMeasurement();
        calibration += measurement.GetCalibration();

        if (i % 3 == 2)
        {
            Board::Display::PutChar(0xff);
        }
    }

    configuration.SetCalibration(measurement.range, calibration / 10);
}


void Application::Calibrate()
{
    Board::Display::Clear();
    Board::Display::PutStringP(PSTR("Kalibracja..."));
    Board::Display::Move(0, 1);

    reader.ReadCalibration();

    while (player.IsPlaying())
    {
        Board::Nop();
        player.Play();
        reader.Execute();
    }

    this->CalibrateRange(Range::_400mV);
    this->CalibrateRange(Range::_4V);
    this->CalibrateRange(Range::_40V);
    this->CalibrateRange(Range::_400V);

    Board::SetTimeout(ConfigWriteTimeout);
    this->configuration.Write();
    Board::SetTimeout(SdInitTimeout);

    measurer.SetRange(configuration.range);
    Board::Display::Clear();
}


void Application::AdcIntHandler()
{
    TIFR0 |= _BV(OCF0A);
    measurer.AddSample(ADC);
}


void Application::Timer1OverflowIntHandler()
{
    int16_t divider = (int16_t) 1 << volumeDivider;
    uint16_t sample = 2048 + player.GetSample() / divider;
    uint16_t cmd = 0x7000 | (sample);

    Board::DacCsPin::SetHigh();
    Board::DacCsPin::SetLow();
    
    Board::DacSpi::Out(cmd >> 8);
    Board::DacSpi::Out(cmd & 0xff);
}


void Application::Timer2OverflowIntHandler()
{
    timeMs += 10;
}